# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *

# ----------------------------------------------------------------------------

# Enforce Python version check during package import.
# This is the same check as the one at the top of setup.py
import sys

__minimum_python_version__ = "3.5"


class UnsupportedPythonError(Exception):
    pass


if sys.version_info < tuple(
    (int(val) for val in __minimum_python_version__.split("."))
):
    raise UnsupportedPythonError(
        "so_pysm_models does not support Python < {}".format(__minimum_python_version__)
    )

if not _ASTROPY_SETUP_:
    # For egg_info test builds to pass, put package imports here.
    pass

from .synchrotron import GaussianSynchrotron
from .dust import GaussianDust
from .alms import PrecomputedAlms
from .co_lines import COLines
from .extragalactic import WebSkySZ, WebSkyCIB, WebSkyCMB, WebSkyCMBMap, WebSkyCMBTensor
from .so_models import get_so_models



import os.path
from pathlib import Path

from numba import njit
import numpy as np

try:  # PySM >= 3.2.1
    import pysm3.units as u
    import pysm3 as pysm
except ImportError:
    import pysm.units as u
    import pysm

from .alms import PrecomputedAlms
from . import utils


@njit
def y2uK_CMB(nu):
    """Compton-y distortion at a given frequency

    Parmeters:
    nu (float): frequency in GHz

    Returns:
    float: intensity variation dT_CMB in micro-Kelvin
      dT_CMB = dI_nu / (dB_nu / dT)_Tcmb
      where B_nu is the Planck function and dI_nu is the intensity distortion

    """

    h = 6.62607004e-27
    k = 1.380622e-16
    Tcmb = 2.725
    x = h * nu * 1e9 / k / Tcmb
    return 1e6 * Tcmb * (x * (np.exp(x) + 1) / (np.exp(x) - 1) - 4)


class WebSkyCIB(pysm.InterpolatingComponent):
    """PySM component interpolating between precomputed maps"""

    def __init__(
        self,
        websky_version="0.3",
        input_units="MJy / sr",
        nside=4096,
        interpolation_kind="linear",
        map_dist=None,
        verbose=False,
        local_folder=None,
        coord="C",
    ):
        self.local_folder = local_folder
        super().__init__(
            str(websky_version),
            input_units,
            nside,
            interpolation_kind,
            map_dist=map_dist,
            verbose=verbose,
        )
        self.remote_data = utils.RemoteData(coord)

    def get_filenames(self, path):
        """Get filenames for a websky version
        For a standard interpolating component, we list files in folder,
        here we need to know the names in advance so that we can only download the required maps
        """

        websky_version = path
        if websky_version == "0.3":

            available_frequencies = []
            for base_freq in [27, 39, 93, 145, 225, 280]:
                for delta_freq in [-1, 0, 1]:
                    available_frequencies.append(base_freq + delta_freq)
            for base_freq in [100, 217, 353, 545, 857]:
                delta_freq = 0
                available_frequencies.append(base_freq + delta_freq)

            filenames = {
                freq: "websky/0.3/{nside}cib_{:04d}.fits".format(
                    freq, nside="512/" if self.nside <= 512 else ""
                )
                for freq in available_frequencies
            }
        if self.local_folder is not None:
            for freq in filenames:
                filenames[freq] = os.path.join(self.local_folder, filenames[freq])

        return filenames

    def read_map_by_frequency(self, freq):
        filename = self.remote_data.get(self.maps[freq])
        return self.read_map_file(freq, filename)


class WebSkySZ(pysm.Model):
    def __init__(
        self,
        version="0.3",
        sz_type="kinetic",
        nside=4096,
        map_dist=None,
        verbose=False,
        coord="C",
    ):

        super().__init__(nside=nside, map_dist=map_dist)
        self.version = str(version)
        self.sz_type = sz_type
        self.verbose = verbose
        self.remote_data = utils.RemoteData(coord)
        filename = self.remote_data.get(self.get_filename())
        self.m = self.read_map(filename, field=0, unit=u.uK_CMB)

    def get_filename(self):
        """Get SZ filenames for a websky version"""

        path = Path("websky") / self.version

        if self.nside <= 512:
            path /= "512"

        if self.sz_type == "kinetic":
            path = path / "ksz.fits"
        elif self.sz_type == "thermal":
            path = path / "tsz.fits"

        return str(path)

    @u.quantity_input
    def get_emission(self, freqs: u.GHz, weights=None) -> u.uK_RJ:

        freqs = pysm.check_freq_input(freqs)
        weights = pysm.normalize_weights(freqs, weights)

        # input map is in uK_CMB, we multiply the weights which are
        # in uK_RJ by the conversion factor of uK_CMB->uK_RJ
        # this is the equivalent of
        weights = (weights * u.uK_CMB).to_value(
            u.uK_RJ, equivalencies=u.cmb_equivalencies(freqs * u.GHz)
        )

        is_thermal = self.sz_type == "thermal"
        output = (
            get_sz_emission_numba(freqs, weights, self.m.value, is_thermal) << u.uK_RJ
        )

        # the output of out is always 2D, (IQU, npix)
        return output


@njit(parallel=True)
def get_sz_emission_numba(freqs, weights, m, is_thermal):
    output = np.zeros((3, len(m)), dtype=m.dtype)
    for i in range(len(freqs)):
        if is_thermal:
            signal = m * m.dtype.type(y2uK_CMB(freqs[i]))
        else:
            signal = m
        pysm.utils.trapz_step_inplace(freqs, weights, i, signal, output[0])
    return output


class WebSkyCMBTensor(PrecomputedAlms):
    def __init__(
        self,
        websky_version,
        nside,
        precompute_output_map=False,
        tensor_to_scalar=1e-3,
        map_dist=None,
        coord="C",
    ):
        """Websky CMB tensor-mode BB component

        Websky-compatible unlensed BB component due to primordial tensor perturbations
        The inputs are simulated with tensor-to-scalar ratio `r` of 1,
        then scaled by the `tensor_to_scalar` input parameter.

        Parameters
        ----------
        websky_version : str
            Websky version, see the documentation for more information
        nside : int
            Desired output HEALPix N_side
        precompute_output_map : bool
            If True, the output map is precomputed in the constructor
        tensor_to_scalar : float
            Tensor to scalar ratio `r`, ratio between the tensor and the
            scalar perturbations power spectra
        map_dist : pysm.MapDist
            see the PySM documentation
        """

        filename = utils.RemoteData(coord).get(
            "websky/{}/tensor_cl_r1_nt0.fits".format(websky_version)
        )
        self.tensor_to_scalar = tensor_to_scalar
        super().__init__(
            filename,
            input_units="uK_CMB",
            input_reference_frequency=None,
            nside=nside,
            precompute_output_map=precompute_output_map,
            has_polarization=True,
            from_cl=True,
            from_cl_seed=0,  # always do same realization
            map_dist=map_dist,
        )

    def compute_output_map(self, alm):
        return super().compute_output_map(alm) * np.sqrt(self.tensor_to_scalar)


class WebSkyCMB(PrecomputedAlms):
    def __init__(
        self,
        websky_version,
        nside,
        precompute_output_map=False,
        seed=1,
        lensed=True,
        map_dist=None,
        coord="C",
    ):
        filename = utils.RemoteData(coord).get(
            "websky/{}/{}lensed_alm_seed{}.fits".format(
                websky_version, "" if lensed else "un", seed
            )
        )
        super().__init__(
            filename,
            input_units="uK_CMB",
            input_reference_frequency=None,
            nside=nside,
            precompute_output_map=precompute_output_map,
            has_polarization=True,
            map_dist=map_dist,
        )


class WebSkyCMBMap(pysm.CMBMap):
    def __init__(
        self,
        websky_version,
        nside,
        precompute_output_map=False,
        seed=1,
        lensed=True,
        include_solar_dipole=False,
        map_dist=None,
        coord="C",
    ):
        template_nside = 512 if nside <= 512 else 4096
        lens = "" if lensed else "un"
        soldip = "solardipole_" if include_solar_dipole else ""
        filenames = [
            utils.RemoteData(coord).get(
                f"websky/{websky_version}/map_{pol}_{lens}lensed_alm_seed{seed}_{soldip}nside{template_nside}.fits"
            )
            for pol in "IQU"
        ]
        super().__init__(
            map_I=filenames[0],
            map_Q=filenames[1],
            map_U=filenames[2],
            nside=nside,
            map_dist=map_dist,
        )



import os.path

import numpy as np
import scipy.stats as stats
import healpy as hp

import pytest
from astropy.tests.helper import assert_quantity_allclose

try:  # PySM >= 3.2.1
    import pysm3.units as u
except ImportError:
    import pysm.units as u

from .. import PrecomputedAlms


@pytest.fixture()
def setup(tmpdir):
    # tmpdir is a py.test feature to provide a temporary folder

    folder = tmpdir.mkdir("alms")

    np.random.seed(12)
    alm_size = hp.Alm.getsize(lmax=100)
    alms = 1j * np.random.normal(size=(3, alm_size)) << u.K_CMB
    alms += np.random.normal(size=(3, alm_size)) << u.K_CMB

    # str needed to support Python 3.5
    filename = os.path.join(str(folder), "alms.fits")
    hp.write_alm(filename, alms)
    return alms, filename


def test_precomputed_alms(setup):

    alms, filename = setup

    nside = 64
    # we assume the original `alms` are in `K_CMB`
    ref_freq = 40 * u.GHz
    test_map_K_CMB = hp.alm2map(alms, nside=nside) << u.K_CMB

    alms_K_RJ = alms.to(u.K_RJ, equivalencies=u.cmb_equivalencies(ref_freq))
    filename_K_RJ = filename.replace(".fits", "_RJ.fits")
    hp.write_alm(filename_K_RJ, alms_K_RJ)

    precomputed_alms = PrecomputedAlms(
        filename=filename_K_RJ,
        nside=nside,
        input_units="K_RJ",
        input_reference_frequency=ref_freq,
    )
    m = precomputed_alms.get_emission(23 * u.GHz)

    assert_quantity_allclose(
        m, test_map_K_CMB.to(u.K_RJ, equivalencies=u.cmb_equivalencies(23 * u.GHz))
    )

    freqs = np.array([1, 10, 100]) * u.GHz

    for freq in freqs:
        np.testing.assert_allclose(
            precomputed_alms.get_emission(freq),
            test_map_K_CMB.to(u.K_RJ, equivalencies=u.cmb_equivalencies(freq)),
        )


def test_precomputed_alms_K_CMB(setup):

    alms, filename = setup

    nside = 64
    test_map = hp.alm2map(alms, nside=nside) << u.K_CMB
    precomputed_alms = PrecomputedAlms(
        filename=filename, nside=nside, input_units="K_CMB"
    )

    freqs = np.array([1, 10, 100]) * u.GHz

    for freq in freqs:
        np.testing.assert_allclose(
            precomputed_alms.get_emission(freq),
            test_map.to(u.K_RJ, equivalencies=u.cmb_equivalencies(freq)),
        )


def test_from_cl(tmpdir):
    nside = 256
    lmax = 512

    folder = tmpdir.mkdir("cls")

    filename = os.path.join(str(folder), "cls.fits")

    input_cl = np.zeros((6, lmax + 1), dtype=np.double)
    # using healpy old ordering TT, TE, TB, EE, EB, BB
    # using healpy new ordering TT, EE, BB, TE, TB, EB
    input_cl[3] = 1e5 * stats.norm.pdf(np.arange(lmax + 1), 250, 30)  # EE
    hp.write_cl(filename, input_cl, overwrite=True)

    precomputed_alms = PrecomputedAlms(
        filename=filename,
        nside=nside,
        input_units="K_CMB",
        from_cl=True,
        from_cl_seed=100,
    )
    freq = 100 * u.GHz
    m = precomputed_alms.get_emission(freq)
    m = m.to(u.K_CMB, equivalencies=u.cmb_equivalencies(freq))

    cl = hp.anafast(m, lmax=lmax)
    # anafast returns results in new ordering
    # TT, EE, BB, TE, EB, TB
    np.testing.assert_allclose(input_cl[3][200:300], cl[1][200:300], rtol=0.2)
    np.testing.assert_allclose(0, cl[0], rtol=1e-3)
    np.testing.assert_allclose(0, cl[2:], rtol=1e-3, atol=1e-4)



import numpy as np
import healpy as hp

def create_high_pass_filter(l1, l2, lmax):
    ell = np.arange(l1, l2)
    wl = np.zeros(lmax+1)
    wl[l2:] = 1.0
    wl[ell] = 0.5 * (1 - np.cos(np.pi * (ell - l1) / (l2 - l1)))
    return wl


def create_low_pass_filter(l1, l2, lmax):
    ell = np.arange(l1, l2)
    wl = np.zeros(lmax+1)
    wl[0:l1] = 1.0
    wl[ell] = 0.5 * (1 - np.cos(np.pi * (l2 - ell) / (l2 - l1)))
    return wl


def apply_filter(hmap, filt):
    nside = hp.get_nside(hmap)
    lmax = 3 * nside
    filt = filt[0:lmax]
    alm = hp.map2alm(hmap, lmax=lmax)
    almf = hp.almxfl(alm, filt)
    hmap_out = hp.alm2map(almf, nside)
    return hmap_out



# Autogenerated by Astropy-affiliated package so_pysm_models's setup.py on 2021-02-12 19:20:29 UTC
from __future__ import unicode_literals
import datetime

version = "2.4.0"
githash = "562e71d2f6405ef50134c696dc8a7a9ee1eb5d20"


major = 2
minor = 4
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2021, 2, 12, 19, 20, 29)
debug = False

astropy_helpers_version = "3.0.2"



import numpy as np

import healpy as hp

try:  # PySM >= 3.2.1
    import pysm3.units as u
    import pysm3 as pysm
except ImportError:
    import pysm.units as u
    import pysm

from .utils import RemoteData


class COLines(pysm.Model):
    def __init__(
        self,
        target_nside,
        output_units,
        has_polarization=True,
        line="10",
        include_high_galactic_latitude_clouds=False,
        polarization_fraction=0.001,
        theta_high_galactic_latitude_deg=20.0,
        random_seed=1234567,
        verbose=False,
        run_mcmole3d=False,
        map_dist=None,
        coord="C",
    ):

        """Class defining attributes for CO line emission.
        CO templates are extracted from Type 1 CO Planck maps.
        See further details in https://www.aanda.org/articles/aa/abs/2014/11/aa21553-13/aa21553-13.html

        Parameters
        ----------
        target_nside : int
            HEALPix NSIDE of the output maps
        output_units : str
            unit string as defined by `pysm.convert_units`, e.g. uK_RJ, K_CMB
        has_polarization : bool
            whether or not to simulate also polarization maps
        line : string
            CO rotational transitions.
            Accepted values : 10, 21, 32
        polarization_fraction: float
            polarisation fraction for polarised CO emission.
        include_high_galactic_latitude_clouds: bool
            If True it includes a simulation from MCMole3D to include
            high Galactic Latitude clouds.
            (See more details at http://giuspugl.github.io/mcmole/index.html)
        run_mcmole3d: bool
            If True it simulates  HGL cluds by running MCMole3D, otherwise it coadds
            a map of HGL emission.
        random_seed: int
            set random seed for mcmole3d simulations.
        theta_high_galactic_latitude_deg : float
            Angle in degree  to identify High Galactic Latitude clouds
            (i.e. clouds whose latitude b is `|b|> theta_high_galactic_latitude_deg`).
        map_dist : mpi4py communicator
            Read inputs across a MPI communicator, see pysm.read_map
        """

        self.line = line
        self.line_index = {"10": 0, "21": 1, "32": 2}[line]
        self.line_frequency = {
            "10": 115.271 * u.GHz,
            "21": 230.538 * u.GHz,
            "32": 345.796 * u.GHz,
        }[line]
        self.target_nside = target_nside

        self.template_nside = 512 if self.target_nside <= 512 else 2048

        super().__init__(nside=target_nside, map_dist=map_dist)

        self.remote_data = RemoteData(coord)

        self.planck_templatemap_filename = "co/HFI_CompMap_CO-Type1_{}_R2.00_ring.fits".format(
            self.template_nside
        )
        self.planck_templatemap = self.read_map(
            self.remote_data.get(self.planck_templatemap_filename),
            field=self.line_index,
            unit=u.K_CMB,
        )

        self.include_high_galactic_latitude_clouds = (
            include_high_galactic_latitude_clouds
        )
        self.has_polarization = has_polarization
        self.polarization_fraction = polarization_fraction
        self.theta_high_galactic_latitude_deg = theta_high_galactic_latitude_deg
        self.random_seed = random_seed
        self.run_mcmole3d = run_mcmole3d

        self.output_units = u.Unit(output_units)
        self.verbose = verbose

    def signal(self):
        """
        Simulate CO signal
        """
        out = (
            hp.ud_grade(map_in=self.planck_templatemap, nside_out=self.target_nside)
            << u.K_CMB
        )

        if self.include_high_galactic_latitude_clouds:
            out += self.simulate_high_galactic_latitude_CO()

        if self.has_polarization:
            Q_map, U_map = self.simulate_polarized_emission(out)
            out = np.array([out, Q_map, U_map])

        convert_to_uK_RJ = (1 * u.K_CMB).to_value(
            self.output_units, equivalencies=u.cmb_equivalencies(self.line_frequency)
        )

        return out * convert_to_uK_RJ

    def simulate_polarized_emission(self, I_map):
        """
        Add polarized emission by means of:
        * an overall constant polarization fraction,
        * a depolarization map to mimick the line of sigth depolarization effect at low Galactic latitudes
        * a polarization angle map coming from a dust template (we exploit the observed correlation
        between polarized dust and molecular emission in star forming regions).
        """
        polangle = self.read_map(
            self.remote_data.get("co/psimap_dust90_{}.fits".format(self.template_nside))
        ).value
        depolmap = self.read_map(
            self.remote_data.get("co/gmap_dust90_{}.fits".format(self.template_nside))
        ).value

        if hp.get_nside(depolmap) != self.target_nside:
            polangle = hp.ud_grade(map_in=polangle, nside_out=self.target_nside)
            depolmap = hp.ud_grade(map_in=depolmap, nside_out=self.target_nside)

        cospolangle = np.cos(2.0 * polangle)
        sinpolangle = np.sin(2.0 * polangle)

        P_map = self.polarization_fraction * depolmap * I_map
        Q_map = P_map * cospolangle
        U_map = P_map * sinpolangle
        return Q_map, U_map

    def simulate_high_galactic_latitude_CO(self):
        """
        Coadd High Galactic Latitude CO emission, simulated with  MCMole3D.
        """
        if self.run_mcmole3d:
            import mcmole3d as cl

            # params to MCMole
            N = 40000
            L_0 = 20.4  # pc
            L_min = 0.3
            L_max = 60.0
            R_ring = 5.8
            sigma_ring = 2.7  # kpc
            R_bulge = 3.0
            R_z = 10  # kpc
            z_0 = 0.1
            Em_0 = 240.0
            R_em = 6.6
            model = "LogSpiral"

            nside = self.target_nside
            Itot_o, _ = cl.integrate_intensity_map(
                self.planck_templatemap,
                hp.get_nside(self.planck_templatemap),
                planck_map=True,
            )
            Pop = cl.Cloud_Population(N, model, randseed=self.random_seed)

            Pop.set_parameters(
                radial_distr=[R_ring, sigma_ring, R_bulge],
                typical_size=L_0,
                size_range=[L_min, L_max],
                thickness_distr=[z_0, R_z],
                emissivity=[Em_0, R_em],
            )
            Pop()

            if self.verbose:
                Pop.print_parameters()
            # project into  Healpix maps
            mapclouds = cl.do_healpy_map(
                Pop,
                nside,
                highgalcut=np.deg2rad(90.0 - self.theta_high_galactic_latitude_deg),
                apodization="gaussian",
                verbose=self.verbose,
            )
            Itot_m, _ = cl.integrate_intensity_map(mapclouds, nside)
            # convert simulated map into the units of the Planck one
            rescaling_factor = Itot_m / Itot_o
            mapclouds /= rescaling_factor
            hglmask = np.zeros_like(mapclouds)
            # Apply mask to low galactic latitudes
            listhgl = hp.query_strip(
                nside,
                np.deg2rad(90.0 + self.theta_high_galactic_latitude_deg),
                np.deg2rad(90 - self.theta_high_galactic_latitude_deg),
            )
            hglmask[listhgl] = 1.0
            rmsplanck = self.planck_templatemap[listhgl].std()
            rmssim = mapclouds[listhgl].std()
            if rmssim == 0.0:
                belowplanck = 1.0
            else:
                belowplanck = rmssim / rmsplanck

            return mapclouds * hglmask / belowplanck
        else:
            mapclouds = self.read_map(
                self.remote_data.get(
                    "co/mcmoleCO_HGL_{}.fits".format(self.template_nside)
                ),
                field=self.line_index,
                unit=u.K_CMB,
            )

            return mapclouds



import errno
import os
import shutil

from distutils.core import Extension
from distutils.ccompiler import get_default_compiler
from distutils.command.build_ext import build_ext as DistutilsBuildExt

from ..distutils_helpers import get_main_package_directory
from ..utils import get_numpy_include_path, import_file

__all__ = ['AstropyHelpersBuildExt']


def should_build_with_cython(previous_cython_version, is_release):
    """
    Returns the previously used Cython version (or 'unknown' if not
    previously built) if Cython should be used to build extension modules from
    pyx files.
    """

    # Only build with Cython if, of course, Cython is installed, we're in a
    # development version (i.e. not release) or the Cython-generated source
    # files haven't been created yet (cython_version == 'unknown'). The latter
    # case can happen even when release is True if checking out a release tag
    # from the repository
    have_cython = False
    try:
        from Cython import __version__ as cython_version  # noqa
        have_cython = True
    except ImportError:
        pass

    if have_cython and (not is_release or previous_cython_version == 'unknown'):
        return cython_version
    else:
        return False


class AstropyHelpersBuildExt(DistutilsBuildExt):
    """
    A custom 'build_ext' command that allows for manipulating some of the C
    extension options at build time.
    """

    _uses_cython = False
    _force_rebuild = False

    def __new__(cls, value, **kwargs):

        # NOTE: we need to wait until AstropyHelpersBuildExt is initialized to
        # import setuptools.command.build_ext because when that package is
        # imported, setuptools tries to import Cython - and if it's not found
        # it will affect the rest of the build process. This is an issue because
        # if we import that module at the top of this one, setup_requires won't
        # have been honored yet, so Cython may not yet be available - and if we
        # import build_ext too soon, it will think Cython is not available even
        # if it is then intalled when setup_requires is processed. To get around
        # this we dynamically create a new class that inherits from the
        # setuptools build_ext, and by this point setup_requires has been
        # processed.

        from setuptools.command.build_ext import build_ext as SetuptoolsBuildExt

        class FinalBuildExt(AstropyHelpersBuildExt, SetuptoolsBuildExt):
            pass

        new_type = type(cls.__name__, (FinalBuildExt,), dict(cls.__dict__))
        obj = SetuptoolsBuildExt.__new__(new_type)
        obj.__init__(value)

        return obj

    def finalize_options(self):

        # First let's find the package folder, then we can check if the
        # version and cython_version are accessible
        self.package_dir = get_main_package_directory(self.distribution)

        version = import_file(os.path.join(self.package_dir, 'version.py'),
                              name='version').version
        self.is_release = 'dev' not in version

        try:
            self.previous_cython_version = import_file(os.path.join(self.package_dir,
                                                                    'cython_version.py'),
                                                       name='cython_version').cython_version
        except (FileNotFoundError, ImportError):
            self.previous_cython_version = 'unknown'

        self._uses_cython = should_build_with_cython(self.previous_cython_version, self.is_release)

        # Add a copy of the _compiler.so module as well, but only if there
        # are in fact C modules to compile (otherwise there's no reason to
        # include a record of the compiler used). Note that self.extensions
        # may not be set yet, but self.distribution.ext_modules is where any
        # extension modules passed to setup() can be found
        extensions = self.distribution.ext_modules
        if extensions:
            build_py = self.get_finalized_command('build_py')
            package_dir = build_py.get_package_dir(self.package_dir)
            src_path = os.path.relpath(
                os.path.join(os.path.dirname(__file__), 'src'))
            shutil.copy(os.path.join(src_path, 'compiler.c'),
                        os.path.join(package_dir, '_compiler.c'))
            ext = Extension(self.package_dir + '.compiler_version',
                            [os.path.join(package_dir, '_compiler.c')])
            extensions.insert(0, ext)

        super().finalize_options()

        # If we are using Cython, then make sure we re-build if the version
        # of Cython that is installed is different from the version last
        # used to generate the C files.
        if self._uses_cython and self._uses_cython != self.previous_cython_version:
            self._force_rebuild = True

        # Regardless of the value of the '--force' option, force a rebuild
        # if the debug flag changed from the last build
        if self._force_rebuild:
            self.force = True

    def run(self):

        # For extensions that require 'numpy' in their include dirs,
        # replace 'numpy' with the actual paths
        np_include = None
        for extension in self.extensions:
            if 'numpy' in extension.include_dirs:
                if np_include is None:
                    np_include = get_numpy_include_path()
                idx = extension.include_dirs.index('numpy')
                extension.include_dirs.insert(idx, np_include)
                extension.include_dirs.remove('numpy')

            self._check_cython_sources(extension)

        # Note that setuptools automatically uses Cython to discover and
        # build extensions if available, so we don't have to explicitly call
        # e.g. cythonize.

        super().run()

        # Update cython_version.py if building with Cython

        if self._uses_cython and self._uses_cython != self.previous_cython_version:
            build_py = self.get_finalized_command('build_py')
            package_dir = build_py.get_package_dir(self.package_dir)
            cython_py = os.path.join(package_dir, 'cython_version.py')
            with open(cython_py, 'w') as f:
                f.write('# Generated file; do not modify\n')
                f.write('cython_version = {0!r}\n'.format(self._uses_cython))

            if os.path.isdir(self.build_lib):
                # The build/lib directory may not exist if the build_py
                # command was not previously run, which may sometimes be
                # the case
                self.copy_file(cython_py,
                               os.path.join(self.build_lib, cython_py),
                               preserve_mode=False)

    def _check_cython_sources(self, extension):
        """
        Where relevant, make sure that the .c files associated with .pyx
        modules are present (if building without Cython installed).
        """

        # Determine the compiler we'll be using
        if self.compiler is None:
            compiler = get_default_compiler()
        else:
            compiler = self.compiler

        # Replace .pyx with C-equivalents, unless c files are missing
        for jdx, src in enumerate(extension.sources):
            base, ext = os.path.splitext(src)
            pyxfn = base + '.pyx'
            cfn = base + '.c'
            cppfn = base + '.cpp'

            if not os.path.isfile(pyxfn):
                continue

            if self._uses_cython:
                extension.sources[jdx] = pyxfn
            else:
                if os.path.isfile(cfn):
                    extension.sources[jdx] = cfn
                elif os.path.isfile(cppfn):
                    extension.sources[jdx] = cppfn
                else:
                    msg = (
                        'Could not find C/C++ file {0}.(c/cpp) for Cython '
                        'file {1} when building extension {2}. Cython '
                        'must be installed to build from a git '
                        'checkout.'.format(base, pyxfn, extension.name))
                    raise IOError(errno.ENOENT, msg, cfn)

            # Cython (at least as of 0.29.2) uses deprecated Numpy API features
            # the use of which produces a few warnings when compiling.
            # These additional flags should squelch those warnings.
            # TODO: Feel free to remove this if/when a Cython update
            # removes use of the deprecated Numpy API
            if compiler == 'unix':
                extension.extra_compile_args.extend([
                    '-Wp,-w', '-Wno-unused-function'])



import numpy as np

import healpy as hp

import math

from . import laws
from . import filter_utils

try:  # PySM >= 3.2.1
    import pysm3.units as u
    import pysm3 as pysm
except ImportError:
    import pysm.units as u
    import pysm


class GaussianDust(pysm.Model):
    def __init__(
        self,
        nside,
        has_polarization=True,
        pixel_indices=None,
        TT_amplitude=350.0,
        Toffset=18.0,
        EE_amplitude=100.0,
        rTE=0.35,
        EtoB=0.5,
        alpha=-0.42,
        beta=1.53,
        temp=19.6,
        nu_0=353,
        seed=None,
        map_dist=None,
    ):
        """Gaussian dust model

        See more details at https://so-pysm-models.readthedocs.io/en/latest/so_pysm_models/models.html

        Parameters
        ----------
        nside : int
            HEALPix NSIDE of the output maps
        has_polarization : bool
            whether or not to simulate also polarization maps
            Default: True
        pixel_indices : ndarray of ints
            Outputs partial maps given HEALPix pixel indices in RING ordering
        TT_amplitude : float
            amplitude of synchrotron TT power spectrum (D_ell) at at the reference
            frequency and ell=80, in muK^2 and thermodinamic units.
            Default: 350. from the amplitude of PySM-d0 dust model at 353GHz
            in the region covered by SO-SAT.
        Toffset : float
            offset to be applied to the temperature map in muK in RJ units.
            Default: 18 from the mean value of the T PySM-s0 synch map at 23GHz
            in the region covered by SO-SAT
        EE_amplitude : float
            Amplitude of EE modes D_ell at reference frequency at ell=80
            Default: 100. from the amplitude of HFI-353 E-modes spectrum in the
            region covered by SO-SAT
        EtoB: float
            ratio between E and B-mode amplitude for dust.
            Default: 0.5 from Planck 2018 IX
        alpha : same as alpha_sync for dust.
            Default: -0.42 from Planck 2018 IX
        beta : float
            dust spectral index.
            Default: 1.53 from Planck 2018 IX
        temp : float
            dust temperature.
            Default: 19.6 from Planck 2018 IX
        nu0 : float
            dust reference frequency in GHz.
            Default: 353
        seed : int
            seed for random realization of map
            Default: None
        """

        super().__init__(nside=nside, map_dist=map_dist)
        self.has_polarization = has_polarization
        self.TT_amplitude = TT_amplitude
        self.Toffset = Toffset
        self.EE_amplitude = EE_amplitude
        self.EtoB = EtoB
        self.alpha = alpha
        self.beta = beta
        self.temp = temp
        self.nu_0 = nu_0
        self.seed = seed

    @u.quantity_input
    def get_emission(self, freqs: u.GHz, weights=None) -> u.uK_RJ:
        """Return map in uK_RJ at given frequency or array of frequencies"""

        nu = pysm.check_freq_input(freqs)
        assert (
            len(nu) == 1
        ), "Bandpass integration not implemented in Gaussian emissions"

        nell = 3 * self.nside
        ell = np.arange(nell)
        dl_prefac = 2 * np.pi / ((ell + 0.01) * (ell + 1))
        clZERO = np.zeros_like(ell)
        clTT_dust = (
            dl_prefac
            * self.TT_amplitude
            * ((ell + 0.1) / 80.0) ** self.alpha
            * laws.black_body_cmb(self.nu_0) ** 2
        )
        clTT_dust[0] = 0.0
        clEE_dust = (
            dl_prefac
            * self.EE_amplitude
            * ((ell + 0.1) / 80.0) ** self.alpha
            * laws.black_body_cmb(self.nu_0) ** 2
        )
        BB_amplitude = self.EE_amplitude * self.EtoB
        clBB_dust = (
            dl_prefac
            * BB_amplitude
            * ((ell + 0.1) / 80.0) ** self.alpha
            * laws.black_body_cmb(self.nu_0) ** 2
        )
        if self.seed == None:
            mseed = np.random.randint(0, 2 ** 32)
        else:
            mseed = self.seed
        np.random.seed(mseed)
        if self.nside <= 64:
            nside_temp = self.nside
        else:
            nside_temp = 64
        amp_dust = np.array(
            hp.synfast(
                [clTT_dust, clEE_dust, clBB_dust, clZERO, clZERO, clZERO],
                nside_temp,
                pol=True,
                new=True,
                verbose=False,
            )
        )
        amp_dust[0] = amp_dust[0] + self.Toffset
        lbreak_TT = 2
        while np.any(amp_dust[0] < 0):
            clTT_dust[1:lbreak_TT] = clTT_dust[lbreak_TT]
            np.random.seed(mseed)
            amp_dust = np.array(
                hp.synfast(
                    [clTT_dust, clEE_dust, clBB_dust, clZERO, clZERO, clZERO],
                    nside_temp,
                    pol=True,
                    new=True,
                    verbose=False,
                )
            )
            amp_dust[0] = amp_dust[0] + self.Toffset
            lbreak_TT += 1
        if self.nside > 64:
            low_pass_filter = filter_utils.create_low_pass_filter(
                l1=30, l2=60, lmax=64 * 3 - 1
            )
            high_pass_filter = filter_utils.create_high_pass_filter(
                l1=30, l2=60, lmax=nell - 1
            )
            clTT_dust_hpf = clTT_dust * high_pass_filter
            amp_dust[0] = filter_utils.apply_filter(amp_dust[0], low_pass_filter)
            np.random.seed(mseed)
            amp_dust_hell = np.array(
                hp.synfast(
                    [clTT_dust_hpf, clEE_dust, clBB_dust, clZERO, clZERO, clZERO],
                    self.nside,
                    pol=True,
                    new=True,
                    verbose=False,
                )
            )
            amp_dust = hp.ud_grade(amp_dust, self.nside)
            amp_dust[1:3] = amp_dust[1:3] * 0.0
            amp_dust += amp_dust_hell
        min_map = np.min(amp_dust[0])
        if min_map < 0:
            Toffset_add = math.ceil(-min_map)
            amp_dust[0] += Toffset_add
        spec_dust = laws.modified_black_body(nu, self.nu_0, self.beta, self.temp)
        out = amp_dust[None, :, :] * spec_dust[:, None, None]
        # the output of out is always 3D, (num_freqs, IQU, npix), if num_freqs is one
        # we return only a 2D array.
        return out[0] << u.uK_RJ



# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2019-01-25 03:21:16 UTC
from __future__ import unicode_literals
import datetime

version = "3.0.2"
githash = "ca630405b54d40695b7bbd2824cb0f142b5811b6"


major = 3
minor = 0
bugfix = 2

release = True
timestamp = datetime.datetime(2019, 1, 25, 3, 21, 16)
debug = False

astropy_helpers_version = ""

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"



try:  # PySM >= 3.2.1
    import pysm3.units as u
    import pysm3 as pysm
except ImportError:
    import pysm.units as u
    import pysm

from .utils import RemoteData

def get_so_models(key, nside, map_dist=None, coord="C"):

    remote_data = RemoteData(coord)

    small_scale = key.endswith("s")
    if small_scale:
        nside_template = 4096
        key = key[:-1]
    else:
        nside_template = 512

    if key.startswith("SO_d"):
        if key == "SO_d0":
            map_mbb_index = 1.53
            map_mbb_temperature = 19.6 * u.K
            dust_T = remote_data.get("dust_T_ns{}.fits".format(nside_template))
            freq_ref_I = 545 * u.GHz
        elif key == "SO_d1":
            map_mbb_index = remote_data.get(
                "variable_spectral_index/beta_dust_ns{}_1deg.fits".format(
                    nside_template
                )
            )
            map_mbb_temperature = remote_data.get(
                "variable_spectral_index/temperature_dust_ns{}_1deg.fits".format(
                    nside_template
                )
            )
            dust_T = remote_data.get("dust_T_ns{}_353GHz.fits".format(nside_template))
            freq_ref_I = 353 * u.GHz
        model = pysm.ModifiedBlackBody(
            dust_T,
            remote_data.get("dust_Q_ns{}.fits".format(nside_template)),
            remote_data.get("dust_U_ns{}.fits".format(nside_template)),
            freq_ref_I=freq_ref_I,
            freq_ref_P=353 * u.GHz,
            map_mbb_index=map_mbb_index,
            map_mbb_temperature=map_mbb_temperature,
            nside=nside,
            has_polarization=True,
            unit_I=u.uK_RJ,
            unit_Q=u.uK_RJ,
            unit_U=u.uK_RJ,
            unit_mbb_temperature=u.K,
            map_dist=map_dist,
        )
    elif key.startswith("SO_s"):
        if key == "SO_s0":
            index = -3.1
            synch_T = remote_data.get("synch_T_ns{}.fits".format(nside_template))
            freq_ref_I = 408 * u.MHz
        elif key == "SO_s1":
            index = remote_data.get(
                "variable_spectral_index/beta_synch_ns{}_1deg.fits".format(
                    nside_template
                )
            )
            synch_T = remote_data.get("synch_T_ns{}_23GHz.fits".format(nside_template))
            freq_ref_I = 23 * u.GHz
        model = pysm.PowerLaw(
            map_I=synch_T,
            map_Q=remote_data.get("synch_Q_ns{}.fits".format(nside_template)),
            map_U=remote_data.get("synch_U_ns{}.fits".format(nside_template)),
            freq_ref_I=freq_ref_I,
            freq_ref_P=23 * u.GHz,
            map_pl_index=index,
            nside=nside,
            unit_I=u.uK_RJ,
            unit_Q=u.uK_RJ,
            unit_U=u.uK_RJ,
            map_dist=map_dist,
        )
    elif key == "SO_f0":
        model = pysm.PowerLaw(
            map_I=remote_data.get("freefree_T_ns{}.fits".format(nside_template)),
            freq_ref_I=30 * u.GHz,
            map_pl_index=-2.14,
            nside=nside,
            unit_I=u.uK_RJ,
            map_dist=map_dist,
        )
    elif key == "SO_a0":
        model = pysm.Sky(
            component_objects=[
                pysm.SpDust(
                    map_I=remote_data.get("ame1_T_ns{}.fits".format(nside_template)),
                    freq_ref_I=22.8 * u.GHz,
                    emissivity=remote_data.get("ame_emissivity.txt"),
                    freq_peak=18.95 * u.GHz,
                    freq_ref_peak=30 * u.GHz,
                    nside=nside,
                    unit_I=u.uK_RJ,
                    map_dist=map_dist,
                ),
                pysm.SpDust(
                    map_I=remote_data.get("ame2_T_ns{}.fits".format(nside_template)),
                    freq_ref_I=41.0 * u.GHz,
                    emissivity=remote_data.get("ame_emissivity.txt"),
                    freq_peak=33.35 * u.GHz,
                    freq_ref_peak=30 * u.GHz,
                    nside=nside,
                    unit_I=u.uK_RJ,
                    map_dist=map_dist,
                ),
            ]
        )
    elif key == "SO_a1":
        model = pysm.Sky(
            component_objects=[
                pysm.SpDustPol(
                    map_I=remote_data.get("ame1_T_ns{}.fits".format(nside_template)),
                    freq_ref_I=22.8 * u.GHz,
                    emissivity=remote_data.get("ame_emissivity.txt"),
                    freq_peak=remote_data.get(
                        "variable_spectral_index/ame_nu0_peak_ns{}_1deg.fits".format(
                            nside_template
                        )
                    ),
                    freq_ref_peak=30 * u.GHz,
                    pol_frac=0.01,
                    angle_Q=remote_data.get("dust_Q_ns{}.fits".format(nside_template)),
                    angle_U=remote_data.get("dust_U_ns{}.fits".format(nside_template)),
                    nside=nside,
                    unit_I=u.uK_RJ,
                    map_dist=map_dist,
                ),
                pysm.SpDustPol(
                    map_I=remote_data.get("ame2_T_ns{}.fits".format(nside_template)),
                    freq_ref_I=41.0 * u.GHz,
                    emissivity=remote_data.get("ame_emissivity.txt"),
                    freq_peak=33.35 * u.GHz,
                    freq_ref_peak=30 * u.GHz,
                    pol_frac=0.01,
                    angle_Q=remote_data.get("dust_Q_ns{}.fits".format(nside_template)),
                    angle_U=remote_data.get("dust_U_ns{}.fits".format(nside_template)),
                    nside=nside,
                    unit_I=u.uK_RJ,
                    map_dist=map_dist,
                ),
            ]
        )

    return model



from astropy.utils.data import get_pkg_data_filename

import numpy as np
import healpy as hp

import pytest

from .. import COLines


@pytest.mark.parametrize("include_high_galactic_latitude_clouds", [False, True])
def test_co(include_high_galactic_latitude_clouds):

    co = COLines(
        target_nside=16,
        output_units="K_CMB",
        has_polarization=True,
        line="10",
        include_high_galactic_latitude_clouds=include_high_galactic_latitude_clouds,
        polarization_fraction=0.001,
        theta_high_galactic_latitude_deg=20.0,
        random_seed=1234567,
        verbose=False,
        run_mcmole3d=False,
        coord="G",
    )

    co_map = co.signal()

    tag = "wHGL" if include_high_galactic_latitude_clouds else "noHGL"
    expected_map_filename = get_pkg_data_filename(
        "data/CO10_TQUmaps_{}_nside16_ring.fits.zip".format(tag)
    )
    expected_co_map = hp.read_map(expected_map_filename, field=(0, 1, 2))

    np.testing.assert_allclose(co_map, expected_co_map, rtol=1e-5)



import numpy as np
import healpy as hp
from astropy.utils.data import get_pkg_data_filename

from .. import GaussianDust
from astropy.tests.helper import assert_quantity_allclose

try:  # PySM >= 3.2.1
    import pysm3.units as u
except ImportError:
    import pysm.units as u


def test_gaussian_dust():
    test_map_filename = get_pkg_data_filename(
        "data/Gaussian_dust_353GHz_uRJ_testmap_ns128.fits.zip"
    )
    test_map = hp.read_map(test_map_filename, (0, 1, 2)) << u.uK_RJ

    nside = hp.npix2nside(len(test_map[0]))
    gaussian_dust = GaussianDust(nside=nside, seed=18)
    m = gaussian_dust.get_emission(353*u.GHz)

    assert_quantity_allclose(m, test_map)



import healpy as hp
import numpy as np

def create_high_pass_filter(l1, l2, lmax):
    ell = np.arange(l1, l2)
    wl = np.zeros(lmax+1)
    wl[l2:] = 1.0
    wl[ell] = 0.5 * (1 - np.cos(np.pi * (ell - l1) / (l2 - l1)))
    return wl


def create_low_pass_filter(l1, l2, lmax):
    ell = np.arange(l1, l2)
    wl = np.zeros(lmax+1)
    wl[0:l1] = 1.0
    wl[ell] = 0.5 * (1 - np.cos(np.pi * (l2 - ell) / (l2 - l1)))
    return wl

def apply_filter(hmap, filt):
    import healpy as hp
    nside = hp.get_nside(hmap)
    print('apply_filter', nside)
    lmax = 3*nside
    filt = filt[0:lmax]
    alm = hp.map2alm(hmap, lmax=lmax)
    almf = np.array([hp.almxfl(alm[i], filt) for i in range(len(alm))])
    hmap_out = hp.alm2map(almf, nside)
    return hmap_out

def add_gaussian_small_scales(map_in, nside_out, pol=False):
    cl_in = hp.anafast(map_in)
    map_in_ns_out = hp.ud_grade(map_in, nside_out)
    map_in_ns_out_smt = hp.smoothing(map_in_ns_out, fwhm=np.radians(180./1000.))
    print('map smoothed')
    ell_to_fit = np.arange(100, 500)
    hpf = create_high_pass_filter(300, 1000, 4*nside_out)
    ell_hell = np.arange(len(hpf))
    if pol==False:
    	cl_T_to_fit  = cl_in[ell_to_fit]
    else:
    	cl_T_to_fit = cl_in[0][ell_to_fit]
    fit_cl_T = np.polyfit(np.log(ell_to_fit), np.log(cl_T_to_fit), 1)
    cl_hell_T = np.exp(fit_cl_T[0]*np.log(ell_hell)+fit_cl_T[1])*hpf
    cl_hell_T[0] = 0
    cl_hell_zero = np.zeros(len(cl_hell_T))
    if pol:
        cl_E_to_fit = cl_in[1][ell_to_fit]
        cl_B_to_fit = cl_in[2][ell_to_fit]
        fit_cl_E = np.polyfit(np.log(ell_to_fit), np.log(cl_E_to_fit), 1)
        fit_cl_B = np.polyfit(np.log(ell_to_fit), np.log(cl_B_to_fit), 1)
        cl_hell_E = np.exp(fit_cl_E[0]*np.log(ell_hell)+fit_cl_E[1])*hpf
        cl_hell_B = np.exp(fit_cl_B[0]*np.log(ell_hell)+fit_cl_B[1])*hpf
        cl_hell_E[0] = 0
        cl_hell_B[0] = 0
        cl_hell = np.array([cl_hell_T, cl_hell_E, cl_hell_B, cl_hell_zero, cl_hell_zero, cl_hell_zero])
        map_ss = hp.synfast(cl_hell, nside_out, lmax=nside_out*3, pol=True, new=True)
    else:
        map_ss = hp.synfast(cl_hell_T, nside_out, lmax=nside_out*3)
    print('map small scales computed')
    map_ss_mod = map_ss*map_in_ns_out_smt
    if pol==False:
    	coeff_T = np.std(map_ss_mod)/np.std(map_ss)
    	map_out_T = map_ss_mod/coeff_T+map_in_ns_out_smt
    else:
    	coeff_T = np.std(map_ss_mod[0])/np.std(map_ss[0])
    	map_out_T = map_ss_mod[0]/coeff_T+map_in_ns_out_smt[0]
    if np.any(map_out_T<0):
        negative_pix = np.where(map_out_T<0)[0]
        print('negative pixels ', len(negative_pix))
        if pol==False:
            map_out_T[negative_pix] = map_in_ns_out[negative_pix]
        else:
            map_out_T[negative_pix] = map_in_ns_out[0][negative_pix]
    if pol:
        coeff_Q = np.std(map_ss_mod[1])/np.std(map_ss[1])
        coeff_U = np.std(map_ss_mod[2])/np.std(map_ss[2])
        coeff_P = (coeff_Q+coeff_U)/2.
        map_out_Q = map_ss_mod[1]/coeff_P+map_in_ns_out_smt[1]
        map_out_U = map_ss_mod[2]/coeff_P+map_in_ns_out_smt[2]
        map_out = np.array([map_out_T, map_out_Q, map_out_U])
    else:
        map_out = map_out_T
    return map_out



import numpy as np
from scipy import stats
import healpy as hp
import pytest
try:  # PySM >= 3.2.1
    import pysm3.units as u
except ImportError:
    import pysm.units as u

from .. import utils
from .. import WebSkyCIB, WebSkySZ, WebSkyCMBTensor


def test_cib(tmp_path):

    nside = 4
    shape = hp.nside2npix(nside)

    path = tmp_path / "websky" / "0.3" / "512"
    path.mkdir(parents=True)
    hp.write_map(path / "cib_0094.fits", np.zeros(shape, dtype=np.float32))
    hp.write_map(path / "cib_0100.fits", np.ones(shape, dtype=np.float32))

    interp = WebSkyCIB(
        "0.3", "uK_RJ", nside, interpolation_kind="linear", local_folder=tmp_path
    )

    interpolated_map = interp.get_emission(97 * u.GHz)
    np.testing.assert_allclose(
        np.interp(97, [94, 100], [0, 1]) * np.ones(shape) * u.uK_RJ,
        interpolated_map[0],
    )
    np.testing.assert_allclose(
        0 * u.uK_RJ,
        interpolated_map[1:],
    )


def test_ksz(tmp_path, monkeypatch):

    monkeypatch.setattr(utils, "PREDEFINED_DATA_FOLDERS", {"C": [str(tmp_path)]})
    nside = 4
    shape = hp.nside2npix(nside)

    path = tmp_path / "websky" / "0.3" / "512"
    path.mkdir(parents=True)
    hp.write_map(path / "ksz.fits", np.ones(shape, dtype=np.float32))

    ksz = WebSkySZ("0.3", sz_type="kinetic", nside=nside)

    ksz_map = ksz.get_emission(100 * u.GHz)
    np.testing.assert_allclose(
        np.ones(ksz_map[0].shape) * 0.7772276 * u.uK_RJ, ksz_map[0], rtol=1e-4
    )
    np.testing.assert_allclose(np.zeros((2, len(ksz_map[0]))) * u.uK_RJ, ksz_map[1:])


def test_tsz(tmp_path, monkeypatch):

    monkeypatch.setattr(utils, "PREDEFINED_DATA_FOLDERS", {"C": [str(tmp_path)]})
    nside = 4
    shape = hp.nside2npix(nside)

    path = tmp_path / "websky" / "0.3" / "512"
    path.mkdir(parents=True)
    hp.write_map(path / "tsz.fits", np.ones(shape, dtype=np.float32) * 1e-6)

    tsz = WebSkySZ("0.3", sz_type="thermal", nside=nside)

    tsz_map = tsz.get_emission(100 * u.GHz)
    np.testing.assert_allclose(
        np.ones(len(tsz_map[0])) * -3.193671 * u.uK_RJ, tsz_map[0], rtol=1e-4
    )
    np.testing.assert_allclose(np.zeros((2, len(tsz_map[0]))) * u.uK_RJ, tsz_map[1:])


@pytest.mark.parametrize("tensor_to_scalar", [1, 1e-3])
def test_cmb_tensor(tmp_path, monkeypatch, tensor_to_scalar):

    monkeypatch.setattr(utils, "PREDEFINED_DATA_FOLDERS", {"C": [str(tmp_path)]})
    nside = 256
    lmax = 512

    path = tmp_path / "websky" / "0.3"
    path.mkdir(parents=True)

    input_cl = np.zeros((6, lmax + 1), dtype=np.double)
    input_cl[1] = 1e5 * stats.norm.pdf(np.arange(lmax + 1), 250, 30)  # EE
    filename = path / "tensor_cl_r1_nt0.fits"

    hp.write_cl(filename, input_cl, overwrite=True)

    cmb_tensor = WebSkyCMBTensor("0.3", nside=nside, tensor_to_scalar=tensor_to_scalar)

    freq = 100 * u.GHz
    cmb_tensor_map = cmb_tensor.get_emission(freq)
    cmb_tensor_map = cmb_tensor_map.to(
        u.uK_CMB, equivalencies=u.cmb_equivalencies(freq)
    )

    cl = hp.anafast(cmb_tensor_map, use_pixel_weights=True, lmax=lmax)
    # anafast returns results in new ordering
    # TT, EE, BB, TE, EB, TB
    np.testing.assert_allclose(
        input_cl[5][200:300] * tensor_to_scalar, cl[2][200:300], rtol=0.2
    )
    np.testing.assert_allclose(0, cl[:2], rtol=1e-3)
    np.testing.assert_allclose(0, cl[3:], rtol=1e-3, atol=1e-4)



import numpy as np
import healpy as hp

from astropy.utils.data import get_pkg_data_filename
try:  # PySM >= 3.2.1
    import pysm3.units as u
except ImportError:
    import pysm.units as u

from .. import PrecomputedAlms
from astropy.tests.helper import assert_quantity_allclose


def test_precomputed_alms():
    alms_filename = get_pkg_data_filename(
        "data/Planck_bestfit_alm_seed_583_lmax_95_K_CMB.fits.zip"
    )
    save_name = get_pkg_data_filename("data/map_nside_32_from_Planck_bestfit_alm_seed_583_K_CMB.fits.zip")
    nside = 32
    # Make an IQU sim
    precomputed_alms = PrecomputedAlms(
        alms_filename, nside=nside, input_units="uK_CMB",
        has_polarization=True,
        #input_reference_frequency=148*u.GHz
    )
    simulated_map = precomputed_alms.get_emission(148*u.GHz).to(u.uK_CMB, equivalencies=u.cmb_equivalencies(148*u.GHz))
    expected_map = hp.read_map(save_name, field=(0, 1, 2)) << u.uK_CMB
    assert simulated_map.shape[0] == 3
    assert_quantity_allclose(simulated_map, expected_map)

def test_precomputed_alms_clip():
    # If clipping to lmax of 95 were not applied, this test would fail
    alms_filename = get_pkg_data_filename(
        "data/Planck_bestfit_alm_seed_583_lmax_120_K_CMB.fits.zip"
    )
    save_name = get_pkg_data_filename("data/map_nside_32_from_Planck_bestfit_alm_seed_583_K_CMB.fits.zip")
    nside = 32
    # Make an IQU sim
    precomputed_alms = PrecomputedAlms(
        alms_filename, nside=nside, input_units="uK_CMB",
        has_polarization=True,
        #input_reference_frequency=148*u.GHz
    )
    simulated_map = precomputed_alms.get_emission(148*u.GHz).to(u.uK_CMB, equivalencies=u.cmb_equivalencies(148*u.GHz))
    expected_map = hp.read_map(save_name, field=(0, 1, 2)) << u.uK_CMB
    assert simulated_map.shape[0] == 3
    assert_quantity_allclose(simulated_map, expected_map)



import numpy as np


def power_law(nu, nu_ref_s, beta_s):
    """ Function to compute power law SED.
    Parameters
    ----------
    nu: float, or array_like(float)
        Frequency in GHz.
    beta_s: float
        Power law index in RJ units.
    Returns
    -------
    array_like(float)
        SED relative to reference frequency.
    """
    x = nu / nu_ref_s
    sed = x ** beta_s
    return sed


def curved_power_law(nu, nu_ref_s, beta_s, beta_c):
    """ Function to compute curved power law SED.
    Parameters
    ----------
    nu: float, or array_like(float)
        Frequency in GHz.
    beta_s: float
        Power law index in RJ units.
    beta_c: float
        Power law index curvature.
    Returns
    -------
    array_like(float)
        SED relative to reference frequency.
    """
    x = nu / nu_ref_s
    sed = x ** (beta_s + beta_c * np.log(nu / nu_ref_s))
    return sed


def black_body_cmb(nu):
    """ Function to compute CMB SED.
    Parameters
    ----------
    nu: float, or array_like(float)
        Frequency in GHz.
    """
    x = 0.0176086761 * nu
    ex = np.exp(x)
    sed = ex * (x / (ex - 1)) ** 2
    return sed


def modified_black_body(nu, nu_ref_d, beta_d, T_d):
    """ Function to compute modified blackbody dust SED.
    Parameters
    ----------
    nu: float or array_like(float)
        Frequency at which to calculate SED.
    nu_ref_d: float
        Reference frequency in GHz.
    beta_d: float
        Power law index of dust opacity.
    T_d: float
        Temperature of the dust.
    Returns
    -------
    array_like(float)
        SED of dust modified black body relative to reference frequency.
    """
    x_to = 0.0479924466 * nu / T_d
    x_from = 0.0479924466 * nu_ref_d / T_d
    sed = (nu / nu_ref_d) ** (1 + beta_d) * (np.exp(x_from) - 1) / (np.exp(x_to) - 1)
    return sed



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This packages contains python packages that are bundled with the package but
are external to it, and hence are developed in a separate source tree. Note
that this package is distinct from the /cextern directory of the source code
distribution, as that directory only contains C extension code.
"""



import numpy as np
import healpy as hp

try:
    from pixell import curvedsky, enmap
except:
    pass

try:  # PySM >= 3.2.1
    import pysm3.units as u
    import pysm3 as pysm
except ImportError:
    import pysm.units as u
    import pysm


class PrecomputedAlms(object):
    def __init__(
        self,
        filename,
        input_units="uK_CMB",
        input_reference_frequency=None,
        nside=None,
        target_shape=None,
        target_wcs=None,
        from_cl=False,
        from_cl_seed=None,
        precompute_output_map=True,
        has_polarization=True,
        map_dist=None,
    ):
        """Generic component based on Precomputed Alms

        Load a set of Alms from a FITS file and generate maps at the requested
        resolution and frequency assuming the CMB black body spectrum.
        A single set of Alms is used for all frequencies requested by PySM,
        consider that PySM expects the output of components to be in uK_RJ.
        See more details at https://so-pysm-models.readthedocs.io/en/latest/so_pysm_models/models.html

        Also note that the Alms are clipped to 3*nside-1 to avoid
        artifacts from high-ell components which cannot be properly represented
        by a low-nside map.

        Parameters
        ----------
        filename : string
            Path to the input Alms in FITS format
        input_units : string
            Input unit strings as defined by pysm.convert_units, e.g. K_CMB, uK_RJ, MJysr
        input_reference_frequency: float
            If input units are K_RJ or Jysr, the reference frequency
        nside : int
            HEALPix NSIDE of the output maps
        from_cl : bool
            If True, the input file contains C_ell instead of a_lm,
            they should provided with the healpy old ordering TT, TE, TB, EE, EB, BB, sorry.
        from_cl_seed : int
            Seed set just before synalm to simulate the alms from the C_ell,
            necessary to set it in order to get the same input map for different runs
            only used if `from_cl` is True
        precompute_output_map : bool
            If True (default), Alms are transformed into a map in the constructor,
            if False, the object only stores the Alms and generate the map at each
            call of the signal method, this is useful to generate maps convolved
            with different beams
        has_polarization : bool
            whether or not to simulate also polarization maps
            Default: True
        """

        self.nside = nside
        self.shape = target_shape
        self.wcs = target_wcs
        self.filename = filename
        self.input_units = u.Unit(input_units)
        self.has_polarization = has_polarization

        if from_cl:
            np.random.seed(from_cl_seed)
            cl = hp.read_cl(self.filename)
            if not self.has_polarization and cl.ndim > 1:
                cl = cl[0]
            # using healpy old ordering TT, TE, TB, EE, EB, BB
            alm = hp.synalm(cl, new=False, verbose=False)
        else:
            alm = np.complex128(
                hp.read_alm(
                    self.filename, hdu=(1, 2, 3) if self.has_polarization else 1
                )
            )

        self.equivalencies = (
            None
            if input_reference_frequency is None
            else u.cmb_equivalencies(input_reference_frequency)
        )
        if precompute_output_map:
            self.output_map = self.compute_output_map(alm)

        else:
            self.alm = alm

    def compute_output_map(self, alm):

        lmax = hp.Alm.getlmax(alm.shape[-1]) # we assume mmax = lmax
        if self.nside is None:
            assert (self.shape is not None) and (self.wcs is not None)
            n_comp = 3 if self.has_polarization else 1
            output_map = enmap.empty((n_comp,) + self.shape[-2:], self.wcs)
            curvedsky.alm2map(alm, output_map, spin=[0, 2], verbose=True)
        elif self.nside is not None:
            if lmax > 3*self.nside-1:
                clip = np.ones(3*self.nside)
                if alm.ndim == 1:
                    alm_clipped = hp.almxfl(alm, clip)
                else:
                    alm_clipped = [hp.almxfl(each, clip) for each in alm]
            else:
                alm_clipped = alm
            output_map = hp.alm2map(alm_clipped, self.nside)
        else:
            raise ValueError("You must specify either nside or both of shape and wcs")
        return (output_map << self.input_units).to(
            u.uK_CMB, equivalencies=self.equivalencies
        )

    @u.quantity_input
    def get_emission(
        self,
        freqs: u.GHz,
        fwhm: [u.arcmin, None] = None,
        weights=None,
        output_units=u.uK_RJ,
    ):
        """Return map in uK_RJ at given frequency or array of frequencies

        Parameters
        ----------
        freqs : list or ndarray
            Frequency or frequencies in GHz at which compute the signal
        fwhm : float (optional)
            Smooth the input alms before computing the signal, this can only be used
            if the class was initialized with `precompute_output_map` to False.
        output_units : str
            Output units, as defined in `pysm.convert_units`, by default this is
            "uK_RJ" as expected by PySM.
        Returns
        -------
        output_maps : ndarray
            Output maps array with the shape (num_freqs, 1 or 3 (I or IQU), npix)
        """

        freqs = pysm.utils.check_freq_input(freqs)
        weights = pysm.utils.normalize_weights(freqs, weights)

        try:
            output_map = self.output_map
        except AttributeError:
            if fwhm is None:
                alm = self.alm
            else:
                alm = hp.smoothalm(
                    self.alm, fwhm=fwhm.to_value(u.radian), pol=True, inplace=False
                )

            output_map = self.compute_output_map(alm)

        output_units = u.Unit(output_units)
        assert output_units in [u.uK_RJ, u.uK_CMB]
        if output_units == u.uK_RJ:

            convert_to_uK_RJ = (
                np.ones(len(freqs), dtype=np.double) * u.uK_CMB
            ).to_value(u.uK_RJ, equivalencies=u.cmb_equivalencies(freqs * u.GHz))

            if len(freqs) == 1:
                scaling_factor = convert_to_uK_RJ[0]
            else:
                scaling_factor = np.trapz(convert_to_uK_RJ * weights, x=freqs)

            return output_map.value * scaling_factor << u.uK_RJ
        elif output_units == output_map.unit:
            return output_map



import os

# If this package has tests data in the tests/data directory, add them to
# the paths here, see commented example
paths = ["coveragerc", os.path.join("data", "*fits.zip")]


def get_package_data():
    return {_ASTROPY_PACKAGE_NAME_ + ".tests": paths}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions.

import os
from astropy.utils import data
import warnings

DATAURL = {
    "C": "https://portal.nersc.gov/project/cmb/so_pysm_models_data/equatorial/",
    "G": "https://portal.nersc.gov/project/cmb/so_pysm_models_data/",
}
PREDEFINED_DATA_FOLDERS = {
    "C": [
        "/global/project/projectdirs/cmb/www/so_pysm_models_data/equatorial",  # NERSC
        "/mnt/sdceph/users/zonca/simonsobs/so_pysm_models_data/equatorial",  # SDSC
    ],
    "G": [
        "/global/project/projectdirs/cmb/www/so_pysm_models_data/",  # NERSC
        "/mnt/sdceph/users/zonca/simonsobs/so_pysm_models_data/",  # SDSC
    ],
}


class RemoteData:
    def __init__(self, coord):
        """Access template from remote server

        PySM input templates are stored on the CMB project space at NERSC
        and are made available via web.
        The get method of this class tries to access data locally from one
        of the PREDEFINED_DATA_FOLDERS defined above, if it fails, it
        retrieves the files and caches them remotely using facilities
        provided by `astropy.utils.data`.

        Parameters
        ----------

        coord : "C" or "G"
            We have templates either in Equatorial or Galactic reference frame
        """
        self.data_url = DATAURL[coord]
        self.data_folders = PREDEFINED_DATA_FOLDERS[coord]

    def get(self, filename):
        for folder in self.data_folders:
            full_path = os.path.join(folder, filename)
            if os.path.exists(full_path):
                return full_path
        with data.conf.set_temp("dataurl", self.data_url), data.conf.set_temp(
            "remote_timeout", 30
        ):
            warnings.warn(f"Retrieve data for {filename} (if not cached already)")
            map_out = data.get_pkg_data_filename(filename, show_progress=True)
        return map_out



import healpy as hp
import sys
from pathlib import Path

if __name__ == "__main__":

    folder = Path(sys.argv[1])

    out_folder = folder / "512"


    for f in folder.glob("[cib,tsz,ksz,kap]*.fits"):
        print(f)
        f = str(f)
        m = hp.read_map(f)
        m_out = hp.ud_grade(m, 512)
        out_file = f.replace(str(folder), str(out_folder))
        print(out_file)
        hp.write_map(out_file, m_out)



#!/usr/bin/env python

# Licensed under a 3-clause BSD style license - see LICENSE.rst

import builtins

# Ensure that astropy-helpers is available
import ah_bootstrap  # noqa

from setuptools import setup
from setuptools.config import read_configuration

from astropy_helpers.setup_helpers import register_commands, get_package_info
from astropy_helpers.version_helpers import generate_version_py

# Store the package name in a built-in variable so it's easy
# to get from other parts of the setup infrastructure
builtins._ASTROPY_PACKAGE_NAME_ = read_configuration('setup.cfg')['metadata']['name']

# Create a dictionary with setup command overrides. Note that this gets
# information about the package (name and version) from the setup.cfg file.
cmdclass = register_commands()

# Freeze build information in version.py. Note that this gets information
# about the package (name and version) from the setup.cfg file.
version = generate_version_py()

# Get configuration information from all of the various subpackages.
# See the docstring for setup_helpers.update_package_files for more
# details.
package_info = get_package_info()

setup(version=version, cmdclass=cmdclass, **package_info)



#!/usr/bin/env python

import argparse
import fnmatch
import os
import sys

#  Recursively generate index.html files for
#  all subdirectories in a directory tree

index_file_name = 'index.html'

CSS = """ <style>

body {
    background: #f4f4f4;
    margin: 2em 1.5em;
}
/*
body {
    margin: 20px;
    background: #f5f5f5;
    -webkit-box-shadow: rgba(89, 89, 89, 0.449219) 2px 1px 9px 0px;
    -moz-box-shadow: rgba(89, 89, 89, 0.449219) 2px 1px 9px 0px;
    box-shadow: rgba(89, 89, 89, 0.449219) 2px 1px 9px 0px;
    border-radius: 11px;
    -moz-border-radius: 11px;
    -webkit-border-radius: 11px;
    height: 100%;
    min-height: 100%;
}
*/
li {
    font-family: sans-serif;
    font-size: 12pt;
    line-height: 14pt;
    list-style:none;
    list-style-type:none;
    padding: 3px 10px;
    margin: 3px 15px;
    display: block;
    clear:both;
}

.content {
    width: 600px;
    background-color: white;
    margin-bottom: 5em;
    padding-bottom: 3em;
    -webkit-box-shadow: rgba(89, 89, 89, 0.449219) 2px 1px 9px 0px;
    -moz-box-shadow: rgba(89, 89, 89, 0.449219) 2px 1px 9px 0px;
    box-shadow: rgba(89, 89, 89, 0.449219) 2px 1px 9px 0px;
    border: 0;
    border-radius: 11px;
    -moz-border-radius: 11px;
    -webkit-border-radius: 11px;
    height: 96%;
    min-height: 90%;
}


.size {
    float: right;
    color: gray;
}

h1 {
    padding: 10px;
    margin: 15px;
    font-size:13pt;
    border-bottom: 1px solid lightgray;
}

a {
    font-weight: 500;
    perspective: 600px;
    perspective-origin: 50% 100%;
    transition: color 0.3s;
    text-decoration: none;
    color: #060606;
}

a:hover,
a:focus {
    color: #e74c3c;
}

a::before {
    background-color: #fff;
    transition: transform 0.2s;
    transition-timing-function: cubic-bezier(0.7,0,0.3,1);
    transform: rotateX(90deg);
    transform-origin: 50% 100%;
}

a:hover::before,
a:focus::before {
    transform: rotateX(0deg);
}

a::after {
    border-bottom: 2px solid #fff;
}

 </style>
"""


def process_dir(top_dir, opts):
    for parentdir, dirs, files in os.walk(str(top_dir)):

        if not opts.dryrun:
            index_file = open(os.path.join(parentdir, index_file_name), "w")
            index_file.write('''<!DOCTYPE html>
    <html>
     <head>{css}</head>
     <body>
      <div class="content">

       <h1>Simons Observatory PySM models files: {curr_dir}/ </h1>

       <li>see <a href="https://github.com/simonsobs/so_pysm_models">github.com/simonsobs/so_pysm_models</a></li>

       <li><a style="display:block; width:100%" href="..">&#x21B0;</a></li>'''.format(
                css=CSS,
                curr_dir=os.path.basename(os.path.abspath(parentdir))
            )
            )

        for dirname in sorted(dirs):

            absolute_dir_path = os.path.join(parentdir, dirname)

            if not os.access(absolute_dir_path, os.W_OK):
                print(("***ERROR*** folder {} is not writable! SKIPPING!".format(absolute_dir_path)))
                continue
            if opts.verbose:
                print(('DIR:{}'.format(absolute_dir_path)))

            if not opts.dryrun:
                index_file.write("""
       <li><a style="display:block; width:100%" href="{link}">&#128193; {link_text}</a></li>""".format(
                    link=dirname,
                    link_text=dirname))

        for filename in sorted(files):

            if opts.filter and not fnmatch.fnmatch(filename, opts.filter):
                if opts.verbose:
                    print(('SKIP: {}/{}'.format(parentdir, filename)))
                continue

            if opts.verbose:
                print(('{}/{}'.format(parentdir, filename)))
            filename_escaped = filename
            filename_utf8 = filename

            if filename.strip().lower() == index_file_name.lower():
                continue

            try:
                size = int(os.path.getsize(os.path.join(parentdir, filename)))

                if not opts.dryrun:
                    index_file.write(
    """
       <li>&#x1f4c4; <a href="{link}">{link_text}</a><span class="size">{size}</span></li>""".format(
                                link=filename_utf8,
                                link_text=filename_escaped,
                                size=pretty_size(size))
                    )

            except Exception as e:
                print(('ERROR writing file name:', e))
                print('filename_utf8:')
                repr(filename_utf8)
                print(('filename_escaped:'), end=' ')
                repr(filename_escaped)

        if not opts.dryrun:
            index_file.write("""
  </div>
 </body>
</html>""")
            index_file.close()


# bytes pretty-printing
UNITS_MAPPING = [
    (1024 ** 5, ' PB'),
    (1024 ** 4, ' TB'),
    (1024 ** 3, ' GB'),
    (1024 ** 2, ' MB'),
    (1024 ** 1, ' KB'),
    (1024 ** 0, (' byte', ' bytes')),
]


def pretty_size(bytes, units=UNITS_MAPPING):
    """Human-readable file sizes.

    ripped from https://pypi.python.org/pypi/hurry.filesize/
    """
    for factor, suffix in units:
        if bytes >= factor:
            break
    amount = int(bytes / factor)

    if isinstance(suffix, tuple):
        singular, multiple = suffix
        if amount == 1:
            suffix = singular
        else:
            suffix = multiple
    return str(amount) + suffix

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='''DESCRIPTION:
    Generate directory index files recursively.
    Start from current dir or from folder passed as first positional argument.
    Optionally filter by file types with --filter "*.py". ''')

    parser.add_argument('top_dir',
                        nargs='?',
                        action='store',
                        help='top folder from which to start generating indexes, '
                             'use current folder if not specified',
                        default=os.getcwd())

    parser.add_argument('--filter', '-f',
                        help='only include files matching glob',
                        required=False)

    parser.add_argument('--verbose', '-v',
                        action='store_true',
                        help='***WARNING: this can take a very long time with complex file tree structures***'
                             ' verbosely list every processed file',
                        required=False)

    parser.add_argument('--dryrun', '-d',
                        action='store_true',
                        help="don't write any files, just simulate the traversal",
                        required=False)

    config = parser.parse_args(sys.argv[1:])


    process_dir(config.top_dir, config)



import numpy as np

try:  # PySM >= 3.2.1
    import pysm3 as pysm
    import pysm3.units as u
except ImportError:
    import pysm.units as u
    import pysm

from .. import get_so_models
from astropy.tests.helper import assert_quantity_allclose
import pytest

components_dict = {comp[0]: comp for comp in ["dust", "synchrotron", "freefree", "ame"]}

# Expected I and Q emission in uK_RJ at pixel 98969
expected = {
    "SO_a1": [2.7826194464298837, 0.025349571461346275],
    "SO_s1": [11.109877, -0.020188138587129197],
    "SO_d1": [789.259705, 3.7430164427988215],
    "SO_a0": [1.4868411735277658, 0.0],
    "SO_f0": [4525.092680419471, 0.0],
    "SO_s0": [11.756732686207846, -0.021363560253508863],
    "SO_d0": [921.4452792523148, 4.369898779380189],
}


@pytest.mark.parametrize(
    "model_tag", ["SO_d0", "SO_s0", "SO_f0", "SO_a0", "SO_d1", "SO_s1", "SO_a1"]
)
def test_get_so_models(model_tag):
    """Test the `get_so_models` function

    Only check that we can access the 512 version of the template
    and that the result has no NaN"""

    sky = pysm.Sky(
        nside=128, component_objects=[get_so_models(model_tag, nside=128, coord="G")]
    )
    emission = sky.get_emission(freq=100 * u.GHz)

    assert not np.any(np.isnan(emission))
    # Compare I and Q at pixel 100
    for IQ in [0, 1]:
        if expected[model_tag][IQ] != 0:
            assert_quantity_allclose(
                emission[IQ][98969], expected[model_tag][IQ] * u.uK_RJ, rtol=1e-4
            )
